/*
   Copyright (C) 1997-2001 Id Software, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 */

#ifndef __TV_CLIENT_H
#define __TV_CLIENT_H

#include "tv_local.h"

qboolean CL_ValidateConfigstring( char *string );
void CL_DeltaEntity( msg_t *msg, frame_t *frame, int newnum, entity_state_t *old, unsigned bits );
void CL_ParsePlayerstate( msg_t *msg, player_state_t *oldstate, player_state_t *state );
void CL_ParseDeltaMatchState( msg_t *msg, frame_t *oldframe, frame_t *newframe );
int CL_ParseEntityBits( msg_t *msg, unsigned *bits );
void CL_ParseDelta( msg_t *msg, entity_state_t *from, entity_state_t *to, int number, unsigned bits );

#endif // __TV_CLIENT_H
